default_driver_name = "Joern"

tool_drivers = {
    "joern": "Joern",
    "ocular": "ShiftLeft Ocular",
    "ng-sast": "ShiftLeft NextGen Analysis",
    "ngsast": "ShiftLeft NextGen Analysis",
}

# URL for viewing reports online
hosted_viewer_uri = "https://sarifviewer.azurewebsites.net"
